@@ -23,12 +23,15 @@
 #include "simfilesystem.h"
 #include "simapplication.h"
 #include "callmanager.h"
-#include <qatutils.h>
+#include "simauth.h"
+#include "aidapplication.h"
+#include "qatutils.h"
 
 #include <qstring.h>
 #include <qbytearray.h>
 #include <qregexp.h>
 #include <qdebug.h>
+#include <qglobal.h>
 
 #define PHONEBOOK_NLENGTH 32
 #define PHONEBOOK_TLENGTH 16
@@ -42,10 +45,10 @@
 
 SimXmlNode::SimXmlNode( const QString& _tag )
 {
-    parent = 0;
-    next = 0;
-    children = 0;
-    attributes = 0;
+    parent = nullptr;
+    next = nullptr;
+    children = nullptr;
+    attributes = nullptr;
     tag = _tag;
 }
 
@@ -70,7 +73,7 @@ SimXmlNode::~SimXmlNode()
 void SimXmlNode::addChild( SimXmlNode *child )
 {
     SimXmlNode *current = children;
-    SimXmlNode *prev = 0;
+    SimXmlNode *prev = nullptr;
     while ( current ) {
         prev = current;
         current = current->next;
@@ -80,7 +83,7 @@ void SimXmlNode::addChild( SimXmlNode *child )
     } else {
         children = child;
     }
-    child->next = 0;
+    child->next = nullptr;
     child->parent = this;
 }
 
@@ -88,7 +91,7 @@ void SimXmlNode::addChild( SimXmlNode *child )
 void SimXmlNode::addAttribute( SimXmlNode *child )
 {
     SimXmlNode *current = attributes;
-    SimXmlNode *prev = 0;
+    SimXmlNode *prev = nullptr;
     while ( current ) {
         prev = current;
         current = current->next;
@@ -98,7 +101,7 @@ void SimXmlNode::addAttribute( SimXmlNode *child )
     } else {
         attributes = child;
     }
-    child->next = 0;
+    child->next = nullptr;
     child->parent = this;
 }
 
@@ -175,7 +178,7 @@ SimState::SimState( SimRules *rules, SimXmlNode& e )
         _name = e.getAttribute( "name" );
     }
     SimXmlNode *n = e.children;
-    while ( n != 0 ) {
+    while ( n != nullptr ) {
         if ( n->tag == "chat" ) {
 
             // Load a chat response definition.
@@ -242,7 +245,7 @@ SimChat::SimChat( SimState *state, SimXmlNode& e )
     deleteSMS = false;
     readSMS = false;
 
-    while ( n != 0 ) {
+    while ( n != nullptr ) {
         if ( n->tag == "command" ) {
             _command = n->contents;
             int w=_command.indexOf(QChar('*'));
@@ -268,8 +271,12 @@ SimChat::SimChat( SimState *state, SimXmlNode& e )
         } else if ( n->tag == "switch" ) {
             switchTo = n->getAttribute( "name" );
         } else if ( n->tag == "set" ) {
-	    variables += n->getAttribute( "name" );
+            QString name = n->getAttribute( "name" );
+	    variables += name;
 	    values += n->getAttribute( "value" );
+            int delay = n->getAttribute( "delay" ).toInt();
+            if (delay)
+                delays[name] = delay;
         } else if ( n->tag == "newcall" ) {
             newCallVar = n->getAttribute( "name" );
         } else if ( n->tag == "forgetcall" ) {
@@ -331,22 +338,36 @@ bool SimChat::command( const QString& cmd )
     for ( int varNum = 0; varNum < variables.size(); ++varNum ) {
     	QString variable = variables[varNum];
 	QString value = values[varNum];
-        if ( value != "*" ) {
+        int delay = delays.value(variable, 0);
+        QString val;
+
+        if ( value == "*" )
+            val = wild;
+        else {
             int index = value.indexOf( "${*}" );
-            if ( index == -1 ) {
-                state()->rules()->setVariable( variable, value );
-            } else {
+            if ( index != -1 ) {
                 if ( wild.length() > 0 && wild[wild.length() - 1] == 0x1A ) {
                     // Strip the terminating ^Z from SMS PDU's.
                     wild = wild.left( wild.length() - 1 );
                 }
-                state()->rules()->setVariable
-                    ( variable, value.left( index ) + wild +
-                    value.mid( index + 4 ) );
-            }
-        } else {
-            state()->rules()->setVariable( variable, wild );
+
+                val = value.left( index ) + wild + value.mid( index + 4 );
+            } else
+                val = value;
         }
+
+        if (delay) {
+            QVariantTimer *timer = new QVariantTimer(this->state()->rules());
+
+            timer->param = QVariant::fromValue(QPairKV(variable, val));
+
+            timer->setSingleShot( true );
+
+            connect(timer, &QTimer::timeout, this->state()->rules(),
+                    &SimRules::delaySetVariable);
+            timer->start( delay );
+        } else
+            state()->rules()->setVariable( variable, val );
     }
 
     // Switch to the new state.
@@ -444,7 +465,7 @@ SimUnsolicited::SimUnsolicited( SimState *state, SimXmlNode& e )
 
     timer = new QTimer( this );
     timer->setSingleShot( true );
-    connect( timer, SIGNAL(timeout()), this, SLOT(timeout()) );
+    connect( timer, &QTimer::timeout, this, &SimUnsolicited::timeout );
 }
 
 
@@ -496,46 +517,46 @@ static bool readXmlFile( SimXmlHandler *handler, const QString& filename )
     return !reader.hasError();
 }
 
-SimRules::SimRules( int fd, QObject *p,  const QString& filename, HardwareManipulatorFactory *hmf )
+SimRules::SimRules(qintptr fd, QObject *p,  const QString& filename, HardwareManipulatorFactory *hmf )
     : QTcpSocket(p)
 {
     setSocketDescriptor(fd);
-    machine = 0;
-    toolkitApp = 0;
+    machine = nullptr;
+    toolkitApp = nullptr;
 
     if (hmf)
-        machine = hmf->create(this, 0);
+        machine = hmf->create(this, nullptr);
 
     if (machine) {
-        connect(machine, SIGNAL(unsolicitedCommand(QString)),
-                this, SLOT(unsolicited(QString)));
-        connect(machine, SIGNAL(command(QString)),
-                this, SLOT(command(QString)));
-        connect(machine, SIGNAL(variableChanged(QString,QString)),
-                this, SLOT(setVariable(QString,QString)));
-        connect(machine, SIGNAL(switchTo(QString)),
-                this, SLOT(switchTo(QString)));
+        connect(machine, &HardwareManipulator::unsolicitedCommand,
+                this, &SimRules::unsolicited);
+        connect(machine, &HardwareManipulator::command,
+                this, &SimRules::command);
+        connect(machine, &HardwareManipulator::variableChanged,
+                this, &SimRules::setVariable);
+        connect(machine, &HardwareManipulator::switchTo,
+                this, &SimRules::switchTo);
     }
 
     _callManager = new CallManager(this);
-    connect( _callManager, SIGNAL(send(QString)),
-             this, SLOT(respond(QString)) );
-    connect( _callManager, SIGNAL(unsolicited(QString)),
-             this, SLOT(unsolicited(QString)) );
-    connect( _callManager, SIGNAL(dialCheck(QString,bool&)),
-             this, SLOT(dialCheck(QString,bool&)) );
+    connect( _callManager, &CallManager::send,
+             this, qOverload<const QString &>( &SimRules::respond ) );
+    connect( _callManager, &CallManager::unsolicited,
+             this, &SimRules::unsolicited );
+    connect( _callManager, &CallManager::dialCheck,
+             this, &SimRules::dialCheck );
 
     if ( machine ) {
-        connect( machine, SIGNAL(startIncomingCall(QString,QString,QString)),
-                 _callManager, SLOT(startIncomingCall(QString,QString,QString)) );
-        connect ( _callManager, SIGNAL( callStatesChanged( QList<CallInfo> * ) ),
-                  machine, SLOT( callManagement( QList<CallInfo> * ) ) );
-        connect ( machine, SIGNAL( stateChangedToAlerting() ), _callManager,
-                SLOT( dialingToAlerting() ) );
-        connect ( machine, SIGNAL( stateChangedToConnected() ), _callManager,
-                SLOT( dialingToConnected() ) );
-        connect ( machine, SIGNAL( stateChangedToHangup( int ) ), _callManager,
-                SLOT( hangupRemote( int ) ) );
+        connect( machine, &HardwareManipulator::startIncomingCall, _callManager,
+                 qOverload<const QString &, const QString &, const QString &>( &CallManager::startIncomingCall ) );
+        connect ( _callManager, &CallManager::callStatesChanged,
+                  machine, &HardwareManipulator::callManagement);
+        connect ( machine, &HardwareManipulator::stateChangedToAlerting, _callManager,
+                &CallManager::dialingToAlerting );
+        connect ( machine, &HardwareManipulator::stateChangedToConnected, _callManager,
+                &CallManager::dialingToConnected );
+        connect ( machine, &HardwareManipulator::stateChangedToHangup, _callManager,
+                &CallManager::hangupRemote );
     }
 
     connect(this,SIGNAL(readyRead()),
@@ -543,12 +564,10 @@ SimRules::SimRules( int fd, QObject *p,  const QString& filename, HardwareManipu
     connect(this,SIGNAL(disconnected()),
         this,SLOT(destruct()));
     // Initialize the local state.
-    currentState = 0;
-    defState = 0;
-    return_error_string = "";
-    return_error_count = 0;
+    currentState = nullptr;
+    defState = nullptr;
     usedCallIds = 0;
-    fileSystem = 0;
+    fileSystem = nullptr;
     useGsm0710 = false;
     currentChannel = 1;
     incomingUsed = 0;
@@ -581,7 +600,7 @@ SimRules::SimRules( int fd, QObject *p,  const QString& filename, HardwareManipu
     // Load the other states, and the start state's name (if specified).
     SimXmlNode *n = handler->documentElement()->children;
     QString start = QString();
-    while ( n != 0 ) {
+    while ( n != nullptr ) {
         if ( n->tag == "state" ) {
 
             // Load a new state definition.
@@ -612,10 +631,20 @@ SimRules::SimRules( int fd, QObject *p,  const QString& filename, HardwareManipu
             // Load a phonebook definition.
             loadPhoneBook( *n );
 
+        } else if ( n->tag == "simauth" ) {
+
+            _simAuth = new SimAuth( this, *n );
+
+        } else if ( n->tag == "application" ) {
+            AidApplication *app = new AidApplication( this, *n );
+            _applications.append(app);
         }
         n = n->next;
     }
 
+    if ( _applications.length() > 0 )
+        _app_wrapper = new AidAppWrapper( this, _applications, _simAuth );
+
     // Clean up the XML reader objects.
     delete handler;
 
@@ -858,25 +887,29 @@ void SimRules::destruct()
         simApps.removeAt( 0 );
 
     delete conformanceApp;
-    conformanceApp = NULL;
+    conformanceApp = nullptr;
     delete defaultToolkitApp;
-    defaultToolkitApp = NULL;
-    toolkitApp = NULL;
+    defaultToolkitApp = nullptr;
+    toolkitApp = nullptr;
 
     if ( getMachine() )
         getMachine()->handleNewApp();
 
     if ( defState )
         delete defState;
-    defState = NULL;
+    defState = nullptr;
+
+    if ( _simAuth )
+        delete _simAuth;
+    _simAuth = nullptr;
 
     if ( _callManager )
         delete _callManager;
-    _callManager = NULL;
+    _callManager = nullptr;
 
     if ( fileSystem )
         delete fileSystem;
-    fileSystem = NULL;
+    fileSystem = nullptr;
 
     if (machine) machine->deleteLater();
     deleteLater();
@@ -936,7 +969,7 @@ SimState *SimRules::state( const QString& name ) const
 
     }
     qWarning() << "Warning: no state called \"" << name << "\" has been defined";
-    return 0;
+    return nullptr;
 }
 
 bool SimRules::simCsimOk( const QByteArray& payload )
@@ -963,16 +996,27 @@ bool SimRules::simCsimOk( const QByteArray& payload )
 
 bool SimRules::simCommand( const QString& cmd )
 {
-    /* Process SIM toolkit begin and end commands by forcing the
-     * app back to the main menu.  */
-    if ( cmd == "AT*TSTB" || cmd == "AT*TSTE" ) {
-        if ( !toolkitApp ) {
+    // 3GPP Terminal Response Command
+    if ( cmd.startsWith("AT+CUSATT=") ) {
+        int start = cmd.indexOf( QChar('=') ) + 1;
+        QByteArray response = QAtUtils::fromHex( cmd.mid(start) );
+        QSimTerminalResponse resp = QSimTerminalResponse::fromPdu( response );
+
+        if ( !toolkitApp || !toolkitApp->response( resp ) )
             respond( "ERROR" );
-            return true;
-        }
+
+        return true;
+    }
+
+    // 3GPP Envelope command
+    if ( cmd.startsWith("AT+CUSATE=") ) {
+        int start = cmd.indexOf( QChar('=') ) + 1;
+        QByteArray envelope = QAtUtils::fromHex( cmd.mid(start) );
+        QSimEnvelope env = QSimEnvelope::fromPdu( envelope );
+
+        if (!toolkitApp || !toolkitApp->envelope( env ) )
+            respond( "ERROR" );
 
-        respond( "OK" );
-        toolkitApp->abort();
         return true;
     }
 
@@ -1101,6 +1145,9 @@ void SimRules::command( const QString& cmd )
     if ( _callManager->command( cmd ) )
         return;
 
+    if (_app_wrapper && _app_wrapper->command( cmd ))
+        return;
+
     // Process SIM toolkit related commands with the current SIM application.
     if ( simCommand( cmd ) )
         return;
@@ -1273,7 +1320,7 @@ void SimRules::loadPhoneBook( SimXmlNode& node )
     }
     SimPhoneBook *pb = phoneBooks[name];
     SimXmlNode *n = node.children;
-    while ( n != 0 ) {
+    while ( n != nullptr ) {
         if ( n->tag == "entry" ) {
             // Load a phone book entry.
             int index = n->getAttribute( "index" ).toInt();
@@ -1332,9 +1379,9 @@ void SimRules::phoneBook( const QString& cmd )
     }
 
     if ( cmd.startsWith( "AT+CPBS=?" ) ) {
-        QStringList names = phoneBooks.keys();
+        const QStringList names = phoneBooks.keys();
         QString response = "+CPBS: (";
-        foreach ( QString name, names ) {
+        for ( const QString &name : names ) {
             if ( response.length() > 8 )
                 response += QChar(',');
             response += "\"" + name + "\"";
@@ -1507,16 +1554,9 @@ SimPhoneBook *SimRules::currentPB() const
     if ( phoneBooks.contains( currentPhoneBook ) )
         return phoneBooks[currentPhoneBook];
     else
-        return 0;
+        return nullptr;
 }
 
-void SimRules::setReturnError( const QString &error, uint repeat )
-{
-    return_error_string = error;
-    return_error_count = repeat;
-}
-
-
 int SimRules::newCall()
 {
     int id;
@@ -1591,15 +1631,8 @@ QString expandEscapes( const QString& data, bool eol )
 void SimRules::respond( const QString& resp, int delay, bool eol )
 {
     QString r = expand( resp );
-    if (return_error_string != "") {
-        r = return_error_string;
-        if (return_error_count > 0)
-            return_error_count--;
-        if (return_error_count == 0)
-            return_error_string = "";
-    }
-
     QByteArray escaped = expandEscapes( r, eol ).toUtf8();
+
     if ( !delay ) {
         writeChatData(escaped.data(), escaped.length());
         flush();
@@ -1615,7 +1648,12 @@ void SimRules::respond( const QString& resp, int delay, bool eol )
 
 void SimRules::proactiveCommandNotify( const QByteArray& cmd )
 {
-    unsolicited( "*TCMD: " + QString::number( cmd.size() ) );
+    unsolicited( "+CUSATP: " + QAtUtils::toHex( cmd ) );
+}
+
+void SimRules::modemHandledCommandNotify( const QByteArray& cmd )
+{
+    unsolicited( "*HCMD: " + QAtUtils::toHex( cmd ) );
 }
 
 void SimRules::callControlEventNotify( const QSimControlEvent& evt )
@@ -1626,7 +1664,7 @@ void SimRules::callControlEventNotify( const QSimControlEvent& evt )
 
 void SimRules::delayTimeout()
 {
-    SimDelayTimer *timer = (SimDelayTimer *)sender();
+    SimDelayTimer *timer = dynamic_cast<SimDelayTimer *>(sender());
     int save = currentChannel;
     currentChannel = timer->channel;
     writeChatData(timer->response.toLatin1().data(), timer->response.length());
@@ -1635,6 +1673,15 @@ void SimRules::delayTimeout()
     timer->deleteLater();
 }
 
+void SimRules::delaySetVariable()
+{
+    QVariantTimer *timer = dynamic_cast<QVariantTimer *>(sender());
+    QPairKV kv = timer->param.value<QPairKV>();
+
+    setVariable( kv.first, kv.second );
+
+    delete timer;
+}
 
 void SimRules::dialCheck( const QString& number, bool& ok )
 {
